/*

 jobload_c

 This is the routine to load a job and set up the argument string
 (but no channels). Job is NOT activated. Returns jobid or -1 if error,
 sets errno and _oserr.

 Original code was based on libc_frkex_c on source disk #3 of C68 release 1.05

 Date		Author		Comments
 ----		------		--------
 01/09/1991	P.J. Taylor	Original.
 15/09/1991	P.J. Taylor	Amended.
 26/02/1998	A.D. Ives	Uses getcwd() etc. to obtain the current
				directory strings instead of global variables.
				Tries defaulting the extension to _obj then
				_exe before giving up!
 01/03/1998	A.D.Ives	Implemented the FindFile() function to
				simplify searching using a chain of
				extensions.
				If the file is not executable, loads
				another copy of the Shell - passing the
				filename as a redirection to stdin plus the
				original command line.
				More reformatting.
 02/03/1998	A.D.Ives	Included ThingLoad() (imported from Jonathon
				Hudson's Qasqade - modified version of his
				TryThing routine). Changed jobload() so that
				the very first thing it tries to do is to
				load an executable thing.
 03/03/1998	A.D.Ives	Added code that records the type of
				extension that was matched, and now only
				attempts to open a secondary shell if a
				non-executable file with an extension of
				_bat was found.

 1.04
 ----
 07/03/1998	A.D.Ives	Added support for Rexx programs. (Now searches
				for a _REX extension as the last option).

 1.05
 ----
 09/03/1998	A.D.Ives	Added flag to determine whether the new or
				old C68 environment passing should be used.
				This allows us to pass the complete
				environment down to new C68 programs, or
				just the three directories for older ones.
				Added a test to check the file that we just
				loaded to see if it looks like a modern C68
				program. If so, the flag is forced on.

 1.06
 ----
 10/03/1998	A.D.Ives	Special processing added to FindFile() to
				cope with .\ prefix, then FindFile() moved
				to misc_c.

*/

#include <sys/types.h>
#include <errno.h>
#include <qdos.h>
#include <debug_h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <things.h>
#include "libpjt.h"

extern char *_VENV ;	/* This SHOULD be there from QDOS.H */


jobid_t ThingLoad(char *thing, char *cmd, jobid_t owner) {

/*

 Load an Executable Thing.

 thing		Thing name
 cmd		Command line
 owner		Owner

 Returns 0 for success, or a QDOS/SMS -ve error code.

*/

    int sts ;
    THING_HEADER_EXEC *t ;
    long vers ;
    THING_LINKAGE *p ;
    struct JOBHEADER *jh ;

    sts= (int) sms_uthg(thing, -1, -1, NULL, NULL, (long *) &vers, &p) ;
    if (sts > 0) {
	t= (THING_HEADER_EXEC *) sts ;
	if (t->thh.thh_type == THT_EXEC) {
	    long sa ;
	    long ha ;
	    void *jb, *stk ;
	    short elen, nlen ;

	    sa= (long) t + t->thh_strt ;
	    ha= (long) t + t->thh_hdrs ;

	    sts= mt_cjob(t->thh_hdrl, t->thh_data + 4, (char *) sa, owner, &jb);
	    if (sts > 0) {
		memcpy(jb, (void *) ha, t->thh_hdrl) ;
		jh= (((struct JOBHEADER *) jb) - 1) ;
		stk= (void *) jh->jb_regs.jb_A7 ;
		if (cmd && (nlen= strlen(cmd))) {
		    elen= (nlen + 1) & ~1 ;
		}
		else {
		    nlen= elen= 0 ;
		}
		if (elen) {
		    ((char *) stk)-= elen ;
		    (void *) jh->jb_regs.jb_A7= stk ;
		}

		*((short *) stk)++ = 0 ;	/* No channels */

		*((short *) stk)++ = nlen ;
		while (nlen--) {
		    *((char *) stk)++ = *cmd++ ;
		}
		sms_fthg(thing, -1, NULL, 0, NULL, NULL) ;
	    }
	}
	else {
	    sts= -15 ;
	}
    }
    else {
	sts= -7 ;
    }
    return sts ;
}

static int fsheadr(chanid_t chan, timeout_t tim, char *buf, int len) {

  struct REGS
       ri ,			/* Input to QDOS trap		*/
       ro ;			/* Output from QDOS trap	*/
  int res;			/* Value to be returned 	*/

  ri.D0= (datareg_t)0x47 ;
  ri.D2= (datareg_t)len ;
  ri.D3= (datareg_t)tim ;
  ri.A0= (addreg_t)chan ;
  ri.A1= (addreg_t)buf ;
  qdos3(&ri, &ro);
  _oserr= (long)ro.D0 ;
  if ( _oserr )
    res= 0  ;
  else
    res= (int)(ro.D1 & 0xFFFF) ;

}

jobid_t jobload(char *name, char *arg, jobid_t owner) {

  /* name	Name of program 		*/
  /* arg	Argument string 		*/
  /* owner	Job to become owner of this job */

  char *jaddr ; 		/* -> Start of created job		*/
  struct qdirect hdrbuf ;
  struct JOBHEADER *jobhdr ;	/* Header for job we will create	*/
  long progsize,		/* Program size 			*/
       datasize ;		/* Static data size			*/
  char *job_a7, 		/* Stack Pointer for parameter set up	*/
       *p,
       *p1 ;
  int PassEnviron ,		/* Flag to indicate environment pointer
				   should be passed			*/
      CmdLen ,			/* Length of actual command line	*/
      DataLen , 		/* Length of Data Directory		*/
      ProgLen , 		/* Length of Program Directory		*/
      DestLen , 		/* Length of Destination Directory	*/
      Exten ;			/* Code indicating matched extension	*/
#define EXT_	  0
#define EXT_BAT   1
#define EXT_OBJ   2
#define EXT_EXE   3
#define EXT_REX   4
  jobid_t new_jobid ;		/* Job ID to return on fork		*/
  long o_oserr ,		/* Place to save _oserr 		*/
       o_errno ;		/* Place to save errno			*/
  short i;
  chanid_t pchan ;		/* QDOS channel opened for program	*/
  char FullName[MAXNAMELEN + 1] ;
  char DataDir[64] ,		/* Holds the Data Directory		*/
       ProgDir[64] ,		/* Holds the Program Directory		*/
       DestDir[64] ;		/* Holds the Destination Directory	*/

  /* Before we try anything else, try and load a Thing */

  if ( (new_jobid= ThingLoad(name, arg, owner)) >= 0 )
    return new_jobid ;

  /* Get local copies of current directories */

  getcwd(DataDir, 64) ;
  getcpd(ProgDir, 64) ;
  getcdd(DestDir, 64) ;

  errno= 0 ;
  _oserr= 0 ;
  PassEnviron= ShellOpt("+e") ;

/*

  Try to open a file, in this order ...

	Name
	ProgDir + Name
	DataDir + Name

  ... then repeat the sequence with the extensions _bat, _obj, _exe
  and finally _rex appended to the name.

*/

  Exten= EXT_ ;
  FindFile(&pchan, name, "", FullName, ProgDir, DataDir) ;
  if ( errno || _oserr ) {
    Exten= EXT_BAT ;
    FindFile(&pchan, name, "_bat", FullName, ProgDir, DataDir) ;
  }
  else {

    /* If we opened the file straight away, identify its extension */

    int ExtenPos ;
    char *ExtenPtr ;

    ExtenPos= strrpos(name, '_') ;
    if (ExtenPos != -1) {
      ExtenPtr= name + ExtenPos ;
      if ( strcmpi(ExtenPtr, "_bat") == 0 )
	Exten= EXT_BAT ;
      else
      if ( stricmp(ExtenPtr, "_obj") == 0 )
	Exten= EXT_OBJ ;
      else
      if ( stricmp(ExtenPtr, "_exe") == 0 )
	Exten= EXT_EXE ;
      else
      if ( stricmp(ExtenPtr, "_rex") == 0 )
	Exten= EXT_REX ;
    }
  }
  if ( errno || _oserr ) {
    Exten= EXT_OBJ ;
    FindFile(&pchan, name, "_obj", FullName, ProgDir, DataDir) ;
  }
  if ( errno || _oserr ) {
    Exten= EXT_EXE ;
    FindFile(&pchan, name, "_exe", FullName, ProgDir, DataDir) ;
  }
  if ( errno || _oserr ) {
    Exten= EXT_REX ;
    FindFile(&pchan, name, "_rex", FullName, ProgDir, DataDir) ;
  }

  /* If we (finally) got to here, we've exhausted all the options */

  if ( errno || _oserr )
    return -1 ;

  /* OK, so we got a file opened, but what IS it? */

  if ( fsheadr((chanid_t)(pchan), (timeout_t)(-1), (char *)(&hdrbuf), sizeof(hdrbuf)) < 14 ) {
    errno= EOSERR ;
  }
 
  if ( !errno ) {
    if ( hdrbuf.d_type != QF_PROG_TYPE ) {      /* The file is not executable */

      char ShellName[MAXNAMELEN + 1], *Ptr ;

      io_close(pchan) ; 	/* Finished with the channel */

      if ( Exten == EXT_BAT ) {

	/* Find out the name of the Shell */

	Ptr= getenv("SHELL") ;
	if ( Ptr == NULL ) {
	  strcpy(ShellName, ProgDir) ;
	  strcat(ShellName, "Sh") ;
	} else
	  strcpy(ShellName, Ptr) ;
      }
      else if ( Exten == EXT_REX ) {

	/* Find out the name of the REXX Interpreter */

	Ptr= getenv("SHELL_REX") ;
	if ( Ptr == NULL ) {
	  strcpy(ShellName, ProgDir) ;
	  strcat(ShellName, "Rexx") ;
	} else
	  strcpy(ShellName, Ptr) ;
      }
      else
	return -1 ;	/* Not executable, Batch or Rexx, report error */

      chanopn(&pchan, ShellName, -1L, OLD_SHARE) ;
      if ( !errno ) {

	/* Build a command line for the secondary Shell */

	char TempArg[512] ;

	if ( Exten == EXT_BAT )
	  strcpy(TempArg, "<") ;	/* Sh needs stdin redirected	*/
	strcat(TempArg, FullName) ;	/* The filename that we found	*/
	strcat(TempArg, " ") ;		/* Need a separator		*/
	strcat(TempArg, arg) ;		/* Then the original args	*/
	strcpy(arg, TempArg) ;		/* Replace the original args	*/

	/* Get the 'real' file header */

	if ( fsheadr((chanid_t)(pchan), (timeout_t)(-1), (char *)(&hdrbuf), sizeof(hdrbuf)) < 14 ) {
	  errno= EOSERR ;
	}
      }
    }
  }

  if ( !errno ) {
   progsize= hdrbuf.d_length ;	/* Length of file	*/
   datasize= 86 ;		/* Assume 20 channels	*/
   CmdLen= strlen( arg ) ;

  /* Add the space needed for the command line (including 2 byte
     field for command line length, and rounded up. */

    datasize+= hdrbuf.d_datalen + ( (CmdLen + 3) & ~1 ) ;

  /* Now add the space for the directories passed */

    datasize+= (DataLen= (strlen(DataDir) + 1)) ;
    datasize+= (ProgLen= (strlen(ProgDir) + 1)) ;
    datasize+= (DestLen= (strlen(DestDir) + 1)) ;

  /* Ensure that it's even (add 2 for the "Magic Number") */

    datasize= (datasize + 3) & ~1 ;

    new_jobid= mt_cjob((long)progsize, (long)datasize, (char *)(0), owner, &jaddr) ;
    if ( _oserr )
      errno= EOSERR ;
  }
  if( !errno ) {
    if ( fs_load(pchan, jaddr, (long)progsize ) < progsize ) {
      o_oserr= _oserr ;
      o_errno= errno ;
      mt_frjob(new_jobid, -1) ;
      _oserr= o_oserr ;
      errno= o_errno ;
    }
  }

  /* If there were any file errors do a tidy exit. */

  if ( errno ) {
    o_oserr= _oserr ;
    o_errno= errno ;
    io_close( pchan );
    errno= o_errno ;
    _oserr= o_oserr ;
    return -1 ;
  }

  io_close(pchan) ;

  /* Place parameters on the Job's stack */

  /* Try and determine whether this is a modern C68 program file and, if
     so, flag that we should pass the complete environment to it */

  PassEnviron= ( PassEnviron || IsNewC68(jaddr) ) ;
  if ( ShellOpt("-e") )
    PassEnviron= FALSE ;

  jobhdr= (struct JOBHEADER *)( jaddr - sizeof( struct JOBHEADER )) ;
  job_a7= (char *)jobhdr->jb_regs.jb_A7 ;	/* Get stack pointer	*/

  if ( PassEnviron ) {
    job_a7-= 4 ;
    *((long *)job_a7) = (long)_VENV ;		/* Environment Pointer	*/
  }
  else {
    if ( (DataLen + ProgLen + DestLen) & 1 )	/* Odd total length	*/
      job_a7-- ;
    job_a7-= DestLen ;
    strcpy(job_a7, DestDir) ;
    job_a7-= ProgLen ;
    strcpy(job_a7, ProgDir) ;
    job_a7-= DataLen ;
    strcpy(job_a7, DataDir) ;
  }
  p1= (job_a7-= 2) ;

  /* Subtract even command line length plus 2 for length */

  p= (job_a7-= ((CmdLen + 3) & ~1)) + 2 ;

  /* p now points at the place to put the first element in the command line,
     job_a7 two bytes before this. */

  strcpy(p, arg) ;
  *(INT16 *) job_a7= CmdLen ;	/* Set up length field */

  /* Magic number used to indicate that directories are passed must be inserted
     after command line was added as adding command line overwrites one byte of
     magic number. */

  if ( PassEnviron )
    *(short *)p1= 0x4AFC ;
  else
    *(short *)p1= 0x4AFB ;

  job_a7-= 2 ;
  *(short *)job_a7= 0 ; 			/* No channels for now	*/

  jobhdr->jb_regs.jb_A7= (addreg_t)job_a7 ;	/* Set Stack Pointer	*/

  return new_jobid ;
}
